# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''UIAutomation API driver
'''

from __future__ import absolute_import, print_function

import base64
import os
import time
import urllib
import uuid

from qt4i.driver.instruments.ins import Instruments
from qt4i.driver.rpc import rpc_method
from qt4i.driver.rpc import RPCEndpoint
from qt4i.driver.tools import mobiledevice
from qt4i.driver.tools.dt import DT
from qt4i.driver.tools.logger import rotate_logger_by_name, get_logger_path_by_name
from qt4i.driver.util.uimap import xctest2instruments


class Device(RPCEndpoint):
    '''UIATarget/UIAApplication API
    '''
    rpc_name_prefix = "device."

    def __init__(self, rpc_server, device_id):
        self.udid = device_id
        RPCEndpoint.__init__(self)
        self._instruments = Instruments(device_id)
        self.rpc_server = rpc_server
    
    @rpc_method
    def stop_agent(self):
        '''空操作(此方法只用于xctest关闭Agent)
        '''
        pass
    
    @rpc_method
    def call_primitive_function(self, method, *params):
        '''执行UIATarget的方法(此方法保留，以期应对苹果升级或调整UIA框架)
        
        :param _method      : 方法名
        :type _method       : str
        :param _params      : 参数列表(参数_method之后的任意个数参数直接转传)
        :type _params       : []
        :returns            : 方法的输出
        '''
        command = self.create_json_command('uia.target.function', method, list(params))
        return self._instruments.exec_command(command)
    
    @rpc_method
    def install(self, ipa_path):
        '''在设备上安装制定路径下的APP
        
        :param ipa_path: 要安装的app的路径
        :type ipa_path: str
        :returns: boolean -- True if successfully installed, False if not
        '''
        dt = DT()
        return dt.install(ipa_path, self.udid)
    
    @rpc_method
    def uninstall(self, bundle_id):
        '''卸载设备上的APP
        
        :param bundle_id: 要卸载的APP的Bundle_ID, 例如"tencent.com.sng.test.gn"
        :type bundle_id: str
        :returns: boolean -- if uninstall app successfully, False if not
        '''
        dt = DT()
        return dt.uninstall(bundle_id, self.udid)
    
    @rpc_method
    def reboot(self):
        '''重启手机
        '''
        mobiledevice.reboot(self.udid)
    
    @rpc_method
    def start_app(self, bundle_id, app_params, env, trace_template=None, trace_output=None, retry=3, timeout=55):
        '''启动设备上的指定APP
        
        :param bundle_id: 要启动的APP的Bundle ID
        :type bundle_id: str
        :param app_params: APP 使用的参数
        :type app_params: dict
        :param trace_template: 专项测试使用trace_template的路径
        :type trace_template: str
        :param trace_output: 专项测试使用trace_output
        :type trace_output: str
        :param retry: 启动失败重试次数
        :type retry: int
        :param timeout: 单次启动超时 (秒)
        :type timeout: int
        :returns: boolean -- True if start successfully, False if not
        ''' 
        addr = self.rpc_server.server_address
        base_uri = "http://"+addr[0]+":"+str(addr[1])
        xmlrpc_uri = '/'.join([base_uri, "device", self.udid])
        return self._instruments.start(bundle_id, app_params, env, xmlrpc_uri, trace_template, trace_output, retry, timeout)
    
    @rpc_method
    def stop_app(self, bundle_id):
        '''停止指定的app
        
        :param bundle_id: 要停止的APP的Bundle ID
        :type bundle_id: str
        :returns: boolean -- True if stop successfully, False if not
        '''
        return self._instruments.release()

    
    @rpc_method
    def get_rect(self):
        '''获取设备屏幕大小
        
        :returns: 坐标和高宽
        '''
        command = self.create_json_command('uia.target.get_rect')
        return self._instruments.exec_command(command)
    
    @rpc_method
    def get_model(self):
        '''获取设备模型
        
        :returns: str
        '''
        command = self.create_json_command('uia.target.get_model')
        return self._instruments.exec_command(command)
    
    @rpc_method
    def get_name(self):
        '''获取设备名
        
        :returns: str
        '''
        command = self.create_json_command('uia.target.get_name')
        return self._instruments.exec_command(command)
    
    @rpc_method
    def get_system_name(self):
        '''获取系统名
        
        :returns : str
        '''
        command = self.create_json_command('uia.target.get_system_name')
        return self._instruments.exec_command(command)
    
    @rpc_method
    def get_system_version(self):
        '''获取系统版本
        
        :returns: str
        '''
        command = self.create_json_command('uia.target.get_system_version')
        return self._instruments.exec_command(command)
    
    @rpc_method
    def capture_screen(self):
        '''截屏
        :returns: str -- base64编码后截图数据
        '''
        try:
            filename = '/tmp/%s.png' % uuid.uuid1()
            command = self.create_json_command('uia.target.capture_screen', filename)
            self._instruments.exec_command(command)
        except:
            device = DT().get_device_by_udid(self.udid)
            if device["simulator"]:
                raise Exception("instruments screenshot error")
            if not mobiledevice.get_screenshot(self.udid, filename):
                raise Exception("mobiledevice screenshot error")
        start_time = time.time()
        while time.time() - start_time < 15:
            if os.path.isfile(filename):
                break
            time.sleep(0.05)
        else:
            raise Exception('capture_screen error')
        
        with open(filename, "rb") as fd:
            base64_img = base64.b64encode(fd.read())
        return base64_img

    @rpc_method
    def get_element_tree(self):
        '''从顶层开始获取UI树（输出的JSON字符串数据需要URL解码）
        
        :returns: dict
        '''
        command = self.create_json_command('uia.target.get_element_tree')
        result = urllib.unquote(self._instruments.exec_command(command))
        result = result.replace('\r', '\\r')
        result = result.replace('\n', '\\n')
        result = result.replace('\\', '\\\\')
        result = eval(result)
        return result
    
    @rpc_method
    def get_element_tree_and_capture_screen(self, filepath=None):
        '''获取UI树并截屏(该接口仅限UISpy使用)
        
        :param filepath: 将图片存储至该路径（png格式）
        :type filepath: str
        :returns: dict : { element_tree: {}, capture_screen: file_path }
        '''
        command = self.create_json_command('uia.target.get_element_tree_and_capture_screen', filepath)
        result_str = urllib.unquote(self._instruments.exec_command(command))
        return result_str

    @rpc_method
    def set_alert_auto_handling_rules(self, rules):
        '''设置自动处理Alert规则
        
                    当message_text匹配到Alert框内的任意文本，则点击button_text指定的按钮。
                    文本以段为单元，元素的label、name、value分为三段文本。规则匹配范围是Alert框内的所有元素的文本。
        :param rules: 规则集合
        :type rules: list - [ {"message_text": "message_text", "button_text": "button_text"}, ... ]
        '''
        command = self.create_json_command('uia.target.set_rules_of_alert_auto_handle', rules)
        self._instruments.exec_command(command)

    @rpc_method
    def get_alert_auto_handling_rules(self):
        '''获取已设置的自动处理Alert规则
        
        :returns: list - [ {"message_text": "message_text", "button_text": "button_text"}, ... ]
        '''
        command = self.create_json_command('uia.target.get_rules_of_alert_auto_handle')
        return self._instruments.exec_command(command)

    @rpc_method
    def add_alert_auto_handling_rule(self, message_text, button_text):
        '''自动处理Alert规则，添加一项规则
        
        :param message_text: Alert内文本片段，支持正则表达式（范围是所谓文本，每个元素的label/name/value为文本段）
        :type message_text: str
        :param button_text: Alert内按钮的文本，支持正则表达式（元素的label/name/value）
        :type button_text: str
        '''
        command = self.create_json_command('uia.target.add_rule_of_alert_auto_handle', message_text, button_text)
        self._instruments.exec_command(command)

    @rpc_method
    def clear_alert_auto_handling_rules(self):
        '''自动处理Alert规则，清空所有规则
        '''
        command = self.create_json_command('uia.target.clean_rules_of_alert_auto_handle')
        self._instruments.exec_command(command)

    @rpc_method
    def disable_auto_alert_handling(self):
        '''关闭自动关闭alert框
        '''
        command = self.create_json_command('uia.target.turn_off_auto_close_alert')
        self._instruments.exec_command(command)
    
    @rpc_method
    def enable_auto_close_handling(self):
        '''打开自动关闭alert框(优先级低于set_alert_auto_handling_rules所设置的策略)
        '''
        command = self.create_json_command('uia.target.turn_on_auto_close_alert')
        self._instruments.exec_command(command)
    
    @rpc_method
    def background_app(self, seconds=3):
        '''将App置于后台一定时间
        
        :param seconds: 秒
        :type seconds: int
        '''
        command = self.create_json_command('uia.target.deactivate_app_ror_duration', seconds)
        self._instruments.exec_command(command)
    
    @rpc_method
    def click(self, x, y, options=None):
        '''单击（全局操作）
        
        :param x: 横向坐标（从左向右，屏幕百分比）
        :type x: float
        :param y: 纵向坐标（从上向下，屏幕百分比）
        :type y: float
        :param options: 屏幕操作扩展参数
        :type options: dict: {
                            : tapCount    : 1, 几次，默认1次
                            : touchCount  : 1, 几点，默认1个点（例如两个手指则为2）
                            : duration    : 0  按住的时间，默认是0
                            }
        '''
        if options == None:
            command = self.create_json_command('uia.target.tap', {'x':x, 'y':y})
        else:
            command = self.create_json_command('uia.target.tap_with_option', {'x':x, 'y':y}, options)
        self._instruments.exec_command(command)
    
    @rpc_method
    def double_click(self, x, y):
        '''双击（全局操作）
        
        :param x: 横向坐标（从左向右，屏幕百分比）
        :type x: float
        :param y: 纵向坐标（从上向下，屏幕百分比）
        :type y: float
        '''
        command = self.create_json_command('uia.target.double_tap', {'x':x, 'y':y})
        self._instruments.exec_command(command)
    
    @rpc_method
    def long_click(self, x, y, seconds=3):
        '''长按
        
        :param x: 横向坐标（从左向右，屏幕百分比）
        :type x: float
        :param y: 纵向坐标（从上向下，屏幕百分比）
        :type y: float
        :param seconds: 秒
        :type seconds: int
        '''
        command = self.create_json_command('uia.target.touch_and_hold', {'x':x, 'y':y}, seconds)
        self._instruments.exec_command(command)
    
    @rpc_method
    def drag(self, x0, y0, x1, y1, duration=0.5):
        '''拖拽（全局操作）
        
        :param x0: 起始横向坐标（从左向右，屏幕百分比）
        :type x0: float
        :param y0: 起始纵向坐标（从上向下，屏幕百分比）
        :type y0: float
        :param x1: 终止横向坐标（从左向右，屏幕百分比）
        :type x1: float
        :param y1: 终止纵向坐标（从上向下，屏幕百分比）
        :type y1: float
        :param duration: 起始坐标按下的时间（秒）
        :type duration: float
        '''
        if duration == 0:
            command = self.create_json_command('uia.target.flick_from_to', {'x':x0, 'y':y0}, {'x':x1, 'y': y1})
        else:
            command = self.create_json_command('uia.target.drag_from_to_for_duration', {'x':x0, 'y':y0}, {'x':x1, 'y': y1}, duration)
        self._instruments.exec_command(command)
    
    @rpc_method
    def pinch(self, x0, y0, x1, y1, opposing=True, seconds=1):
        '''捏合/展开
        
        :param x0: 起始横向坐标（从左向右，屏幕百分比）
        :type x0: float
        :param y0: 起始纵向坐标（从上向下，屏幕百分比）
        :type y0: float
        :param x1: 终止横向坐标（从左向右，屏幕百分比）
        :type x1: float
        :param y1: 终止纵向坐标（从上向下，屏幕百分比）
        :type y1: float
        :param opposing: True为展开；False为捏合
        :type opposing: boolean
        :param seconds: 秒
        :type seconds: int
        '''
        if opposing:
            command = self.create_json_command('uia.target.pinch_open_from_to_for_duration', {'x':x0, 'y':y0}, {'x':x1, 'y':y1}, seconds)
        else:
            command = self.create_json_command('uia.target.pinch_close_from_to_for_duration', {'x':x0, 'y':y0}, {'x':x1, 'y':y1}, seconds)
        self._instruments.exec_command(command)
    
    
    @rpc_method
    def rotate(self, x, y, options=None):
        '''旋转
        
        :param x: 横向坐标（从左向右，屏幕百分比）
        :type x: float
        :param y: 纵向坐标（从上向下，屏幕百分比）
        :type y: float
        :param options: 自定义项
        :type options: dict : {
                            : duration    : 1,    持续的时间（秒）
                            : radius      : 50,   半径
                            : rotation    : 100,  圆周，默认是圆周率PI
                            : touchCount  : 2     触摸点数量 2 - 5
                            }
        '''
        command = self.create_json_command('uia.target.rotate_with_options', {'x':x, 'y':y}, options)
        self._instruments.exec_command(command)
    
    @rpc_method
    def set_location(self, coordinates, options=None):
        '''设置设备的GPS坐标
        
        :param coordinates: 经纬度
        :type coordinates: dict : {
                            : latitude    : 100, // 以度为单位的纬度。正值表示纬度赤道以北。负值表示纬度赤道以南。
                            : longitude   : 100  // 以度为单位的经度。测量是相对于零子午线，用正值向东延伸的经络及经络的负值西侧延伸。
                            }
        :param options: 自定义项
        :type options: options : {
                        : altitude            : 50, // 海拔高度，以米为单位，相对于海平面。正值表示海拔高度。负值表示低于海平面的高度。
                        : horizontalAccuracy  : 10, // 水平半径，不确位置时的范围内，以米为单位。负的值是无效的。
                        : verticalAccuracy    : 10, // 垂直半径，不确位置时的范围内，以米为单位。负的值是无效的。
                        : course              : 1,  // 设备是移动的，不确定移动方向
                        : speed               : 1   // 移动速度（米/秒）
                        }
        '''
        if options == None:
            command = self.create_json_command('uia.target.set_location', coordinates)
        else:
            command = self.create_json_command('uia.target.set_location_with_options', coordinates, options)
        self._instruments.exec_command(command)
    
    @rpc_method
    def shake(self):
        '''摇晃设备
        '''
        command = self.create_json_command('uia.target.shake')
        self._instruments.exec_command(command)
    
    @rpc_method
    def unlock(self):
        '''解锁设备
        '''
        command = self.create_json_command('uia.target.unlock')
        self._instruments.exec_command(command)
    
    @rpc_method
    def lock(self, seconds=0):
        '''锁定设备
        :param seconds: 秒, 0表示锁定设备；大于0表示锁定一段时间后解锁
        :type seconds: int
        '''
        if seconds == 0:
            command = self.create_json_command('uia.target.lock')
        else:
            command = self.create_json_command('uia.target.lock_for_duration', seconds)
        self._instruments.exec_command(command)
        
    @rpc_method
    def send_keys(self, keys):
        '''键盘输入
        
        :param keys        : 键入的字符串
        :type keys         : str
        '''
        command = self.create_json_command('uia.keyboard.sent_keys', keys)
        self._instruments.exec_command(command)
    
    @rpc_method
    def get_driver_log(self):
        driver_log_path = self._instruments.get_driver_log()
        log = None
        if driver_log_path and os.path.isfile(driver_log_path):
            with open(driver_log_path, "rb") as fd:
                log = fd.read().decode("utf-8")
        return log
        
    @rpc_method
    def get_crash_log(self, procname):
        '''通过设备的udid的获取指定进程的最新的crash日志
        
        :param proc_name: app的进程名，可通过xcode查看
        :type proc_name: str        
        :returns: str
        '''
        return DT().get_crash_log(self.udid, procname)
    
    @rpc_method
    def get_log(self):
        '''获取日志信息
        '''
        log = ''
        log_path = get_logger_path_by_name('driverserver_%s' % self.udid)
        if log_path and os.path.isfile(log_path):
            with open(log_path, "rb") as fd:
                log = fd.read().decode("utf-8")
        return log
    
    @rpc_method
    def cleanup_log(self):
        '''清理日志
        '''
        rotate_logger_by_name('driverserver_%s' % self.udid)
    
    @rpc_method
    def push_file(self, bundle_id, localpath, remotepath):
        '''拷贝Mac本地文件到手机中sandbox的指定目录地
        
        :param bundle_id: app的bundle id
        :type bundle_id: str   
        :param localpath: Mac上的文件路径
        :type localpath: str
        :param remotepath: sandbox上的目录或者文件，例如：/Library/Caches/test/
        :type remotepath: str        
        :returns: boolean
        '''
        return DT().push_file(self.udid, bundle_id, localpath, remotepath)

    @rpc_method
    def pull_file(self, bundle_id, remotepath, localpath='/tmp', is_dir=False, is_delete = False):
        '''拷贝手机中sandbox指定目录的文件到Mac本地
        
        :param bundle_id: app的bundle id
        :type bundle_id: str   
        :param remotepath: sandbox上的目录或者文件，例如：/Library/Caches/test/
        :type remotepath: str        
        :param localpath: 本地的目录
        :type localpath: str
        :param device_udid: 设备的udid
        :type device_udid: str
        :param is_dir: remotepath是否为目录，默认为单个文件
        :type is_dir: bool
        :param is_delete: 是否删除
        :type is_delete: bool
        :returns: list or None
        '''        
        return DT().pull_file(self.udid, bundle_id, remotepath, localpath, is_dir, is_delete)    

    @rpc_method
    def remove_files(self, bundle_id, file_path):
        '''删除手机上指定app中的文件或者目录
        
        :param bundle_id: app的bundle id
        :type bundle_id: str 
        :param file_path: 待删除的文件或者目录，例如: /Documents/test.log
        :type file_path: str  
        '''
        return DT().remove_files(self.udid, bundle_id, file_path) 
       
    @rpc_method
    def get_screen_orientation(self):
        command = self.create_json_command('uia.application.get_interface_orientation')
        return self._instruments.exec_command(command)
    @rpc_method
    def stop_all_agents(self):    
        pass           

class Application(RPCEndpoint):
    '''UIAApplication API
    '''
    rpc_name_prefix = "app."
    
    def __init__(self, rpc_server, device_id):
        self.udid = device_id
        RPCEndpoint.__init__(self)
        self._instruments = Instruments(device_id)
    
    @rpc_method
    def call_primitive_function(self, method, *params):
        '''执行UIAApplication的方法(此方法保留，以期应对苹果升级或调整UIA框架)
        
        :param _method: 方法名
        :type _method: str
        :param _params: 参数列表(参数_method之后的任意个数参数直接转传)
        :type _params: []
        :returns: 方法的输出
        '''
        command = self.create_json_command('uia.application.function', method, list(params))
        return self._instruments.exec_command(command)

    @rpc_method
    def get_app_bundle_id(self):
        '''获取当前App的BundleID
        
        :returns: str -- App的Bundle ID
        '''
        command = self.create_json_command('uia.application.get_app_bundle_id')
        return self._instruments.exec_command(command)
    
    @rpc_method
    def get_app_version(self):
        '''获取App的Version
        
        :returns: str
        '''
        command = self.create_json_command('uia.application.get_app_version')
        return self._instruments.exec_command(command)

    @rpc_method
    def has_crashed(self): 
        '''当前App是否发生crash
        
        :returns: boolean
        '''
        return self._instruments.get_crash_flag()
    

class Element(RPCEndpoint):
    '''UIAElement API
    '''
    rpc_name_prefix = "element."
    
    def __init__(self, rpc_server, device_id):
        self.udid = device_id
        RPCEndpoint.__init__(self)
        self._instruments = Instruments(device_id)
        
    def decode_find_result(self, result):
        '''js 返回值解码
        
        :returns : dict -- 解码结果
        '''
        path              = urllib.unquote(result.get('path', ''))
        valid_path_part   = urllib.unquote(result.get('valid_path_part', ''))
        invalid_path_part = urllib.unquote(result.get('invalid_path_part', ''))
        elements          = result.get('elements', [])
        for i, item in enumerate(elements): 
            elements[i]['attributes'] = urllib.unquote(item.get('attributes'))
        result.update({'path':path, 'valid_path_part':valid_path_part, 'invalid_path_part':invalid_path_part, 'elements':elements})
        return result
    
    @rpc_method
    def call_primitive_function(self, method, *params):
        '''执行UIAElement的方法(此方法保留，以期应对苹果升级或调整UIA框架)
        :param method      : 方法名
        :type method       : str
        :param params      : 参数列表(参数_method之后的任意个数参数直接转传)
        :type params       : []
        :returns: 方法的输出
        '''
        command = self.create_json_command('uia.element.function', method, list(params))
        return self._instruments.exec_command(command)
    
    @rpc_method
    def find_element(self, locator, timeout=3, interval=0.005, strategy='qpath', parent_id=None):
        '''查找单个element对象，返回查找到的element对象的缓存id，不指定parent_id则从Application下开始搜索。
        
        :param locator: 定位element的字符串，例如: "/classname='UIAWindow'"
        :type locator: str
        :param timeout: 查找element的超时值（单位: 秒）
        :param interval: 在查找element的超时范围内，循环的间隔（单位: 秒）
        :type interval: float
        :param strategy: locator的类型，例如: "qpath"
        :type strategy: str
        :param parent_id: 父element的id，如不指定则从UIAApplication下查找
        :type parent_id: str
        :returns: dict : {'element': id, 'path': <encode str>, 'valid_path_part': <encode str>, 'invalid_path_part': <encode str>, 'find_count': <int>, "find_time": int}
        '''
        if strategy == "id":
            locator = "/name='%s' & maxdepth=20" % locator
            strategy = "qpath"
        locator = xctest2instruments(locator)
        command = self.create_json_command('uia.element.find', locator, timeout, interval, strategy, parent_id)
        result = self._instruments.exec_command(command)
        result = self.decode_find_result(result)
        return result
    
    @rpc_method
    def find_elements(self, locator, timeout=3, interval=0.005, strategy='qpath', parent_id=None):
        '''查找单个element对象，返回查找到的element对象的缓存id。不指定parent_id则从UIAApplication下开始搜索。
        
        :param locator: 定位element的字符串，例如: "/classname='UIAWindow'"
        :type locator: str
        :param timeout: 查找element的超时值（单位: 秒）
        :type timeout: float
        :param interval: 在查找element的超时范围内，循环的间隔（单位: 秒）
        :type interval: float
        :param strategy: locator的类型，例如: "qpath"
        :type strategy: str
        :param parent_id: 父element的id，如不指定则从UIAApplication下查找
        :type parent_id: str
        :returns: list : {'elements': [{'element':id, 'attributes':<encode str>}], 'path': <encode str>, 'valid_path_part': <encode str>, 'invalid_path_part': <encode str>, 'find_count': <int>, "find_time": int}
        '''
        if strategy == "id":
            locator = "/name='%s' & maxdepth=20" % locator
            strategy = "qpath"
        start_time = time.time()
        result = None
        find_count = 0
        find_time = 0
        locator = xctest2instruments(locator)
        while True:
            before_step_time = time.time()
            command = self.create_json_command('uia.element.find_elements', locator, 0.5, interval, strategy, parent_id)
            result = self._instruments.exec_command(command)
            result = self.decode_find_result(result)
            after_step_time = time.time()
            step_time = after_step_time - before_step_time
            find_count += result['find_count']
            find_time += step_time
            if result['elements'] or after_step_time - start_time >= timeout:
                break
            if interval > 0:
                sleep_time = interval - step_time
                if sleep_time > 0:
                    time.sleep(sleep_time)
        result['find_count'] = find_count
        result['find_time'] = find_time
        return result

    @rpc_method
    def find_element_with_predicate(self, element_id, predicate):
        '''通过predicate文本获取第一个匹配的子element(UIAutomation API专用)
        
        :param element_id: element的id
        :type element_id: int
        :param predicate: 预期子element的predicate （例如：“name beginswith 'xxx'”）
        :type predicate: str
        :returns: int
        '''
        command = self.create_json_command('uia.element.first_with_predicate', element_id, predicate)
        return self._instruments.exec_command(command)

    @rpc_method
    def find_element_with_value_for_key(self, element_id, key, value):
        '''通过匹配指定key的value，获取第一个匹配的子element
        
        :param element_id: element的id
        :type element_id: int
        :param key: key （例如：label、name、value）
        :type key: str
        :param value: 对应key的value值
        :type value: str
        :returns: int
        '''
        command = self.create_json_command('uia.element.first_with_value_for_key', element_id, key, value)
        return self._instruments.exec_command(command)
    
    @rpc_method
    def find_elements_with_value_for_key(self, element_id, key, value):
        '''通过匹配指定key的value，获取第一个匹配的子element
        
        :param element_id: element的id
        :type element_id: int
        :param key: key （例如：label、name、value）
        :type key: str
        :param value: 对应key的value值
        :type value: str
        :returns: int
        '''
        command = self.create_json_command('uia.element.with_value_for_key', element_id, key, value)
        return self._instruments.exec_command(command)
    
    @rpc_method
    def get_element_tree(self, element_id, max_depth=0):
        '''获取指定element的所有子孙element树
        
        :param element_id: element的id
        :type element_id: int
        :param max_depth: 树的深度, 默认为0，获取所有子孙控件树
        :type max_depth: int
        :returns: dict
        '''
        command = self.create_json_command('uia.element.get_element_tree', element_id)
        result = urllib.unquote(self._instruments.exec_command(command))
        result = result.replace('\r', '\\r')
        result = result.replace('\n', '\\n')
        result = eval(result)
        return result
    
    @rpc_method
    def get_parent_element(self, element_id):
        '''获取指定element的父element的id
        
        :param element_id: element的id
        :type element_id: int
        :returns: int
        '''
        command = self.create_json_command('uia.element.get_parent', element_id)
        return self._instruments.exec_command(command)
    
    @rpc_method
    def get_children_elements(self, element_id):
        '''获取指定element的子elements集合
        
        :param element_id: element的id
        :type element_id: int
        :returns: list : [int, ...]
        '''
        command = self.create_json_command('uia.element.get_children', element_id)
        return self._instruments.exec_command(command)
    
    @rpc_method
    def get_rect(self, element_id):
        '''获取指定element的坐标和长宽
        
        :param element_id: element的id
        :type element_id: int
        :returns: dict : 坐标和长宽
        '''
        command = self.create_json_command('uia.element.get_rect', element_id)
        return self._instruments.exec_command(command)
    
    @rpc_method
    def capture(self, element_id, filepath=None):
        '''截图（截取指定element的图片，并将图片输出至指定的路径）
        
        :param element_id: element的id
        :type element_id: int
        :param filepath: 将图片存储至该路径（png格式），不传路径 或 路径仅含文件名，则截图后存储至默认路径。
        :type filepath: str
        :returns: str
        '''
        command = self.create_json_command('uia.element.capture', element_id, filepath)
        return self._instruments.exec_command(command)
    
    @rpc_method
    def click(self, element_id, x=None, y=None, options=None):
        '''点击指定的element
        
        :param element_id: element的id
        :type element_id: int
        :param x: x坐标（相对于当前element），可不传入该参数
        :type x: float
        :param y: y坐标（相对于当前element），可不传入该参数
        :type y: float
        '''
        if options == None:
            command = self.create_json_command('uia.element.tap', element_id, x, y)
        else:
            command = self.create_json_command('uia.element.tap_with_options', element_id, options)
        self._instruments.exec_command(command)
            
    
    @rpc_method
    def double_click(self, element_id, x=None, y=None):
        '''双击指定的element
        
        :param element_id: element的id
        :type element_id: int
        :param x: x坐标（相对于当前element），可不传入该参数
        :type x: float
        :param y: y坐标（相对于当前element），可不传入该参数
        :type y: float
        '''
        command = self.create_json_command('uia.element.double_tap', element_id, x, y)
        self._instruments.exec_command(command)
    
    @rpc_method
    def drag(self, element_id,  from_x, from_y, to_x, to_y, duration):
        '''拂去/拖拽等自定义操作
        
        :param element_id: element的id
        :type element_id: int
        :param from_x: 起点 x偏移百分比（从左至右为0.0至1.0）
        :type from_x: float
        :param from_y: 起点 y偏移百分比（从上至下为0.0至1.0）
        :type from_y: float
        :param to_x: 终点 x偏移百分比（从左至右为0.0至1.0）
        :type to_x: float
        :param to_y: 终点 y偏移百分比（从上至下为0.0至1.0）
        :type to_y: float
        :param touchCount: 触摸点（iPhone最多4个手指，iPad可以五个手指）
        :type touchCount: int
        :param duration: 持续时间（秒）
        :type duration: float
        :param repeat: 重复该操作
        :type repeat: int
        :param interval: 重复该操作的间隙时间（秒）
        :type interval: float
        :param options : 自定义项
        :type options : dict : {
                            : touchCount   : 1,                    // 触摸点
                            : duration     : 0.5,                  // 时间
                            : startOffset  : { x: 0.0, y: 0.1},    // 偏移百分比
                            : endOffset    : { x: 1.0, y: 0.1 },   // 偏移百分比
                            : repeat       : 1                     // 重复该操作
                            : interval     : 0                     // 重复该操作的间隙时间（秒）
                            }
        '''
        options = {"touchCount":1, "duration":duration, "startOffset":{"x":from_x, "y":from_y}, \
                    "endOffset":{"x":to_x, "y":to_y}, "repeat":1, "interval":0}
        if duration == 0:
            command = self.create_json_command('uia.element.flick_inside_with_options', element_id, options)
        else:    
            command = self.create_json_command('uia.element.drag_inside_with_options', element_id, options)
        self._instruments.exec_command(command)
    
    @rpc_method
    def rotate(self, element_id, options):
        '''旋转
        
        :param element_id: element的id
        :type element_id: int
        :param _options: 自定义项
        :type _options: dict : {
                            : centerOffset : {x:0.0, y:0.0},   // 中心点
                            : duration     : 1.5,              // 持续时间
                            : radius       : 100,              // 半径
                            : rotation     : 100,              // 旋转弧度的长度，默认为圆周率PI
                            : touchCount   : 2                 // 触摸点（最大5个点，这里默认2个）
                            }
        '''
        command = self.create_json_command('uia.element.rotate_with_options', element_id, options)
        self._instruments.exec_command(command)
    
    @rpc_method
    def scroll_to_visible(self, element_id):
        '''自动滚动到该element为可见
        
        :param element_id: element的id
        :type element_id: int
        '''
        command = self.create_json_command('uia.element.scroll_to_visible', element_id)
        self._instruments.exec_command(command)
    
    @rpc_method
    def long_click(self, element_id, seconds=3):
        '''持续按住
        
        :param element_id: element的id
        :type element_id: int
        :param seconds: 秒
        :type seconds: int
        '''
        command = self.create_json_command('uia.element.touch_and_hold', element_id, seconds)
        self._instruments.exec_command(command)
    
    @rpc_method
    def wait_for_invalid(self, element_id, seconds=5):
        '''等待当前element直到无效
        
        :param element_id: element的id
        :type element_id: int
        :param seconds: 秒
        :type seconds: int
        :returns: boolean -- True if invalid, False if not
        '''
        command = self.create_json_command('uia.element.wait_for_invalid', element_id, seconds)
        return self._instruments.exec_command(command)
    
    @rpc_method
    def set_value(self, element_id, value):
        '''设置指定element的value
        
        :param element_id: element的id
        :type element_id: int
        :param value: 值
        :type value: str
        '''
        command = self.create_json_command('uia.element.set_value', element_id, value)
        self._instruments.exec_command(command)
    
    @rpc_method
    def drag_to_value(self, element_id, value):
        '''slider滑到指定位置(0-1之间)
        :param element_id: element的id
        :type element_id: int
        :param value: 值
        :type value: str
        '''
        command = self.create_json_command('uia.element.drag_to_value', element_id, value)
        self._instruments.exec_command(command)
        
    @rpc_method
    def send_keys(self, element_id, keys):
        '''在指定element中键盘输入
        :param element_id: element的id
        :type element_id: int
        :param keys: 键入的字符串
        :type keys: str
        '''
        command = self.create_json_command('uia.element.sent_keys', element_id, keys)
        self._instruments.exec_command(command)
    
    @rpc_method
    def get_element_attrs(self, element_id):
        '''获取指定element的属性信息
        
        :param element_id: element的id
        :type element_id: int
        :returns: str: dict
        '''
        command = self.create_json_command('uia.element.get_element_dict', element_id)
        result = urllib.unquote(self._instruments.exec_command(command))
        result = result.replace('\r', '\\r')
        result = result.replace('\n', '\\n')
        result = eval(result)
        return result
    
    @rpc_method
    def get_element_attr(self, element_id, attr_name):
        '''获取指定element的属性信息
        
        :param element_id: element的id
        :type element_id: int
        :param attr_name: 属性名
        :type attr_name: str
        :returns: str
        '''
        command = self.create_json_command('uia.element.get_attr', element_id, attr_name)
        return self._instruments.exec_command(command) 
    
    @rpc_method
    def find_elements_with_predicate(self, element_id, predicate):
        '''通过predicate文本获取所有匹配的子element(UIAutomation API专用)
        
        :param element_id: element的id
        :type element_id: int
        :param predicate: 预期子element的predicate （例如：“name beginswith 'xxx'”）
        :type predicate: str
        :returns: list - [id1, id2, id3...]
        '''
        command = self.create_json_command('uia.element.with_predicate', element_id, predicate)
        return self._instruments.exec_command(command)
    



# -*- coding:utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''iOS WebView
'''

from __future__ import absolute_import, print_function, division


import time
import traceback

from qt4i.icontrols import Element
from qt4i.icontrols import Window
from qt4i.qpath import QPath
from qt4w.util import JavaScriptError
from qt4w.webdriver.webkitwebdriver import WebkitWebDriver


class IOSWebDriver(WebkitWebDriver):
    '''iOS WebKit WebDrvier
    '''
    
    driver_script = WebkitWebDriver.driver_script + r'''
    window['qt4w_driver_lib']['getScale'] = function(){return 1;};
    '''


class IOSWebView(Element):
    '''
    iOS 内嵌WebView控件
    '''
    

    def __init__(self, root, locator, title=None, url=None,  **ext):
        Element.__init__(self, root, locator, **ext)
        self._scrollview = root
        self._device = self._app.device
        self._driver = self._device.driver
        self.bundle_id = self._app._bundle_id
        self.udid = self._device.udid
        self.is_simulator = self._device.simulator
        self.title = title
        self.url_key = url
        self._top_offset = 0
        self.page_id = self._init_page_id()
        print("webview init page id:%s" % self.page_id)
            
    def _init_page_id(self):
        retry = 5
        for _ in range(retry):
            try:
                return self._driver.web.get_page_id(self.bundle_id, self.title, self.url_key)
            except Exception:
                err = traceback.format_exc()
                time.sleep(1)
        raise RuntimeError('Failed to init page id reason: %s'%err)

    @property
    def top_offset(self):
        return self._top_offset

    @top_offset.setter
    def top_offset(self, offset):
        self._top_offset = offset

    @property
    def webdriver_class(self):
        '''WebView对应的WebDriver类
        '''
        return IOSWebDriver
    
    @property
    def rect(self):
        '''WebView控件的坐标
        '''
        root_rect = Window.rect.fget(self)   #webview相对native主窗口的坐标 # @UndefinedVariable
        return [root_rect.left, root_rect.top, root_rect.width, root_rect.height]
    
    @property
    def visible_rect(self):
        '''WebView控件可见区域的坐标信息
        '''
        rect = self._scrollview.rect  # webview相对native主窗口的可视区域的坐标
        visible_rect = [rect.left, rect.top, rect.width, rect.height]
        return visible_rect 
    
    def _coordinate_conversion(self, x, y, rel=True):
        '''WebView的坐标转换为native的坐标
        ''' 
        v_rect = self.visible_rect
        if x < 0 or x > v_rect[2] or \
           y < 0 or y > v_rect[3]:
            raise RuntimeError('[x:%s, y:%s]不在可见区域' % (x, y))
        x_abs = x + v_rect[0] + self.rect[0]
        y_abs = y + v_rect[1] + self.rect[1] + self._top_offset
        if rel:
            dev_rect = self._device.rect
            x_rel = float('%.2f' % (x_abs / float(dev_rect.width)))
            y_rel = float('%.2f' % (y_abs / float(dev_rect.height)))
            return (x_rel, y_rel)
        else:
            return (x_abs, y_abs)
    
    def click(self, x_offset, y_offset):
        coordinate = self._coordinate_conversion(x_offset, y_offset)
        self._device.click(coordinate[0], coordinate[1])
        
    def long_click(self, x_offset, y_offset, duration=2):
        coordinate = self._coordinate_conversion(x_offset, y_offset)
        self._device.long_click(coordinate[0], coordinate[1], duration)
     
    def double_click(self, x_offset, y_offset):
        coordinate = self._coordinate_conversion(x_offset, y_offset)
        self._device.double_click(coordinate[0], coordinate[1])
    
    def drag(self, x1, y1, x2, y2):
        coordinate1 = self._coordinate_conversion(x1, y1)
        coordinate2 = self._coordinate_conversion(x2, y2)
        self._device.drag(coordinate1[0], coordinate1[1], coordinate2[0], coordinate2[1])
          
    def eval_script(self, frame_xpaths, script):
        '''javascript脚本注入接口
        
        :param frame_xpaths: 保留，暂不使用
        :type frame_xpaths: str|None
        :param script: javascript脚本
        :type script: str
        '''
        result = None
        try:
            result = self._driver.web.eval_script(self.bundle_id, frame_xpaths, self.page_id, script)
        except Exception:
            err = traceback.format_exc()
            if 'WIPageUpdateError' in err:
                self.page_id = self._driver.web.get_page_id(self.bundle_id, self.title, self.url_key)
                return self.eval_script(frame_xpaths, script)
            else:
                raise
        if 'className' in result and (result['className'] == 'ReferenceError' or result['className'] == 'Error'):
            raise JavaScriptError(frame_xpaths, result['description'])
        elif 'type' in result and 'value' in result:
            return result['value']
        elif 'className' in result and result['className'] == 'Array' and 'preview' in result:
            if 'properties' in result['preview']:
                filter_result = []
                for p in result['preview']['properties']:
                    if p['type'] == 'number':
                        filter_result.append(int(p['value']))
                    else:
                        filter_result.append(p['value'])
                return filter_result    
            else:
                return result['preview']
        else:
            return result

    def send_keys(self, keys):
        self._driver.device.send_keys(keys)


class QT4iBrowserWin(Window):
    '''
    浏览器窗口基类
    '''
    def __init__(self, app):
        self._app = app
        Window.__init__(self, self._app)
        self.scroll_win = Element(self._app, 'WebView')
        locators = {
            'webview': {'type': IOSWebView, 'root': self.scroll_win, 'locator': QPath("/classname='WebView'")},
        }
        self.updateLocator(locators)

    @property
    def webview(self):
        '''
        WebView对象
        '''
        return self.Controls['webview']



# -*- coding:utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''iOS设备模块
'''

from __future__ import absolute_import, print_function, division

import base64
import datetime
import os
import pkg_resources
import subprocess
import sys
import time
import uuid
import socket
import traceback
import six
from six import PY3
from six import PY2

from testbase import context
from testbase.conf import settings
from testbase.resource import LocalResourceHandler
from testbase.resource import LocalResourceManagerBackend
from testbase.resource import TestResourceManager
from testbase import logger
from qt4i.util import EnumDirect, Rectangle

from qt4i.driver.rpc import RPCClientProxy
from qt4i.driver.util import Process


QT4i_LOGS_PATH = os.path.abspath('_attachments')
if not os.path.exists(QT4i_LOGS_PATH):
    os.makedirs(QT4i_LOGS_PATH)
    
Encoding = 'UTF-8'
DEFAULT_ADDR = '127.0.0.1'
DEFAULT_PORT = 12306
DEFAULT_AGENT_PORT = 8100
DEFAULT_ALERT_RULE =  settings.get('QT4I_ALERT_RULES', [{'button_text':'^确定$|^好$|^允许$|^OK$|^Allow$'}])


class DeviceServer(object):
    '''设备服务器
    '''
    def __init__(self, addr, port, udid=None, agent_port=DEFAULT_AGENT_PORT, driver_type=settings.get('IOS_DRIVER', 'xctest'), endpoint_clss=''):
        '''初始化
        
        :param addr: 设备主机ip
        :type addr: str
        :param port: 设备主机端口号
        :type port: int
        :param udid: 设备UDID
        :type udid: str or None
        :param agent_port: agent端口号
        :type agent_port: int
        :param driver_type: driver类型，xctest or instruments
        :type driver_type: str
        :param endpoint_clss: 注册远程调用类集，多个以";"间隔
        :type endpoint_clss: str
        '''
        unzip_path = pkg_resources.resource_filename("qt4i", "driver")  # @UndefinedVariable
        self._script = os.path.join(unzip_path,'driverserver.py')
        self._addr = addr
        self._port = port
        self._udid = udid
        self._agent_port = agent_port
        self._driver_type = driver_type
        self._endpoint_clss = endpoint_clss
        self._python = "python3" if PY3 else "python"
        if self._udid:
            self.pidfile = '/tmp/driverserver_%s.pid' % self._udid
        else:
            self.pidfile = '/tmp/driverserver.pid'
        if hasattr(settings, 'QT4i_ENDPOINT_CLSS'): # 兼容小写配置项
            self._endpoint_clss = ','.join([self._endpoint_clss, settings.QT4i_ENDPOINT_CLSS])
        elif hasattr(settings, 'QT4I_ENDPOINT_CLSS'):
            self._endpoint_clss = ','.join([self._endpoint_clss, settings.QT4I_ENDPOINT_CLSS])
        
    def _wait_for_started(self, timeout=60):
        begin_time = time.time()
        while time.time() - begin_time <= timeout:
            try:
                server = RPCClientProxy('/'.join(["http://%s:%s" % (self._addr, self._port), 'host/']), allow_none=True, encoding=Encoding)
                if server.echo():
                    print('server is ready')
                    return True
            except:
                print('server is not ready...')
            time.sleep(0.2)
        raise Exception('DeviceServer failed to start.')
    
    def start(self):
        '''启动
        '''

        if self._udid:
            if self._endpoint_clss:
                cmd = '%s %s -t %s --pidfile %s -p %d -u %s -a %d -r %s start' % (self._python,
                                                                                  self._script,
                                                                                  self._driver_type,
                                                                                  self.pidfile,
                                                                                  self._port,
                                                                                  self._udid,
                                                                                  self._agent_port,
                                                                                  self._endpoint_clss)
            else:
                cmd = '%s %s -t %s --pidfile %s -p %d -u %s -a %d start' % (self._python,
                                                                            self._script,
                                                                            self._driver_type,
                                                                            self.pidfile,
                                                                            self._port,
                                                                            self._udid,
                                                                            self._agent_port)
        else:
            if self._endpoint_clss:
                cmd = '%s %s -t %s -p %d -r %s start' % (self._python, self._script, self._driver_type, self._port, self._endpoint_clss)
            else:
                cmd = '%s %s -t %s -p %d start' % (self._python, self._script, self._driver_type, self._port)
                print(cmd)
                
        subprocess.Popen(cmd, shell=True)
        self._wait_for_started()
        
    def stop(self):
        '''停止
        '''
        if self._udid:
            subprocess.Popen('%s %s --pidfile %s -p %d -u %s stop' % (self._python, self._script, self.pidfile, self._port, self._udid), shell=True)
        else:
            subprocess.Popen('%s %s --pidfile %s -p %d stop' % (self._python, self._script, self.pidfile, self._port), shell=True)
        
    def restart(self):
        '''重启
        '''
        if self._udid:
            if self._endpoint_clss:
                cmd = '%s %s -t %s --pidfile %s -p %d -u %s -a %d -r %s restart' % (self._python,
                                                                                    self._script,
                                                                                    self._driver_type,
                                                                                    self.pidfile,
                                                                                    self._port,
                                                                                    self._udid,
                                                                                    self._agent_port,
                                                                                    self._endpoint_clss)
            else:
                cmd = '%s %s -t %s --pidfile %s -p %d -u %s -a %d restart' % (self._python,
                                                                              self._script,
                                                                              self._driver_type,
                                                                              self.pidfile,
                                                                              self._port,
                                                                              self._udid,
                                                                              self._agent_port)
        else:
            if self._endpoint_clss:
                cmd = '%s %s -t %s -p %d -r %s restart' % (self._python,
                                                           self._script,
                                                           self._driver_type,
                                                           self._port,
                                                           self._endpoint_clss)
            else:
                cmd = '%s %s -t %s -p %d restart' % (self._python,
                                                     self._script,
                                                     self._driver_type,
                                                     self._port)
            
        subprocess.Popen(cmd, shell=True)
        self._wait_for_started()
 
    def debug(self):
        '''调试模式启动driverserver
        '''
        if self._udid:
            if self._endpoint_clss:
                cmd = '%s %s -t %s --pidfile %s -p %d -u %s -a %d -r %s debug' % (self._python,
                                                                                  self._script,
                                                                                  self._driver_type,
                                                                                  self.pidfile,
                                                                                  self._port,
                                                                                  self._udid,
                                                                                  self._agent_port,
                                                                                  self._endpoint_clss)
            else:
                cmd = '%s %s -t %s --pidfile %s -p %d -u %s -a %d debug' % (self._python,
                                                                            self._script,
                                                                            self._driver_type,
                                                                            self.pidfile,
                                                                            self._port,
                                                                            self._udid,
                                                                            self._agent_port)
        else:
            if self._endpoint_clss:
                cmd = '%s %s -t %s -p %d -r %s debug' % (self._python,
                                                         self._script,
                                                         self._driver_type,
                                                         self._port,
                                                         self._endpoint_clss)
            else:
                cmd = '%s %s -t %s -p %d debug' % (self._python, self._script, self._driver_type, self._port)

        subprocess.Popen(cmd, shell=True)
        self._wait_for_started()
        
    def get_pid(self):
        """
        Returns the PID from pidfile
        """
        try:
            if self.pidfile and os.path.exists(self.pidfile):
                with open(self.pidfile, 'r') as pf:
                    pid = int(pf.read().strip())
            else:
                pid = None
        except (IOError, TypeError):
            pid = None
        return pid
    
    def exist(self):
        '''driverserver进程是否存在
        '''
        pid = self.get_pid()
        if pid:
            if Process().get_process_by_pid(pid, 'driverserver.py'):
                return True
        return False

        
class DeviceResource(object):
    '''设备资源
    '''
    
    def __init__(self, host, port, udid, is_simulator, name, version, csst_uri=None, resource_id=None):
        if host == socket.gethostbyname(socket.gethostname()):
            host = DEFAULT_ADDR
        self._host = host
        self._port = port
        self._udid = udid
        self._name = name
        self._version = version
        if csst_uri:
            self._ws_uri = "ws://%s/proxy/control-%s" % (csst_uri, udid)
        else:
            self._ws_uri = None
        self._driver_url = 'http://%s:%d' % (host, port)
        self._is_sim = is_simulator
        self._resource_id = resource_id

    def __getitem__(self, key):
        if key == 'id':
            return self._resource_id
        else:
            return getattr(self, key)
    
    @property
    def driver_url(self):
        return self._driver_url

    @property
    def ws_uri(self):
        return self._ws_uri

    @property
    def resource_id(self):
        return self._resource_id
    
    @property
    def host(self):
        return self._host
    
    @property
    def udid(self):
        return self._udid
    
    @property
    def is_simulator(self):
        return self._is_sim
    
    @property
    def name(self):
        return self._name
    
    @property
    def version(self):
        return self._version
    
    def __eq__(self, other):
        '''通过IP+UDID判断是否为同一个设备
        '''
        return self._host == other._host and self._udid == other._udid
    
    def __str__(self):
        return "udid:%s, ip:%s" % (self._udid, self._host)


class IOSDeviceResourceHandler(LocalResourceHandler):
    '''iOS设备资源的本地管理
    '''

    def iter_resource(self, res_group=None, condition=None):
        DeviceManager.update_local_devices()
        if isinstance(condition, six.string_types):
            condition = {'udid': condition}
        for dev in DeviceManager.devices:
            if condition:
                for k in condition:
                    if condition[k] == dev[k]:
                        break
                else:
                    if condition:
                        continue
            yield dev

LocalResourceManagerBackend.register_resource_type('ios', IOSDeviceResourceHandler())


class DeviceManager(object):
    '''设备管理类，用于多个设备的申请、查询和管理（此类仅用于Device内部，测试用例中请勿直接使用）
    '''
    
    devices = []


    @staticmethod
    def get_devices_from_driver(driver_ip, driver_port=12306):
        driver = RPCClientProxy('http://%s:%d/host/' % (driver_ip, driver_port), allow_none=True, encoding=Encoding)
        driver_devices = driver.list_devices()
        devices = []
        index = 0
        for dev in driver_devices:
            dev['id'] = dev['udid']
            dev['host'] = driver_ip
            dev['port'] = driver_port
            dev['is_simulator'] = dev['simulator']
            dev.pop('simulator', None)
            dev['version'] = dev['ios']
            dev.pop('ios', None)
            if dev['is_simulator']:
                devices.append(dev)
            else:
                devices.insert(index, dev)
                index += 1
        return devices

    @staticmethod  
    def update_local_devices():
        '''更新本地设备列表
        '''
        devices = []
        if sys.platform == 'darwin':
            DeviceServer(DEFAULT_ADDR, DEFAULT_PORT).start()
            devices.extend(DeviceManager.get_devices_from_driver(DEFAULT_ADDR))
        if hasattr(settings, 'QT4I_REMOTE_DRIVERS'):
            remote_drivers = settings.QT4I_REMOTE_DRIVERS
            for driver in remote_drivers:
                if driver['ip'] == '127.0.0.1':
                    continue
                devices.extend(DeviceManager.get_devices_from_driver(driver['ip'], driver['port']))
        DeviceManager.devices = devices
            

class Device(object):
    '''iOS设备基类（包含基于设备的UI操作接口）
    '''
    Devices = []
    
    @classmethod
    def release_all(cls):
        devices = cls.Devices[:]
        for device in devices:
            device.release()
        del cls.Devices[:]
    
    @classmethod
    def cleanup_all_log(cls):
        for device in cls.Devices:
            device.cleanup_log()
    
    def __init__(self, attrs={}, devicemanager=None):
        '''Device构造函数
        
        :param attrs: 设备UDID字符串｜设备属性字典
                                    keys: udid            - 设备UDID
                                          host            - 设备主机ip
                                          is_simulator    - 是否为模拟器
        :type attrs: str|dict
        :param devicemanager: 设备管理类
        :type devicemanager: DeviceManager
        '''
        cond = {}
        if isinstance(attrs, six.string_types):
            cond['udid'] = attrs
        elif isinstance(attrs, dict):
            cond = attrs.copy()
        else:
            raise Exception('Device attributes type error: %s' % type(attrs))
        ct = None
        if devicemanager:
            self._device_resource = devicemanager.acquire_device(cond)
        else:
            ct = context.current_testcase()
            if ct is None:
                self._test_resources = TestResourceManager(LocalResourceManagerBackend()).create_session()
            else:
                self._test_resources = ct.test_resources
            self._device_resource = self._test_resources.acquire_resource("ios", condition=cond)

        if self._device_resource is None:
            raise Exception('无可用的真机和模拟器: %s' % str(cond))

        props = self._device_resource
        if 'properties' in self._device_resource: # online mode
            devprops = self._device_resource['properties']
            props = { p['name'].encode(Encoding): p['value'].encode(Encoding) for p in devprops }
            props['id'] = self._device_resource['id']

        if 'csst_uri' not in props or props['csst_uri'] == 'None':
            props['csst_uri'] = None

        self._device_resource = DeviceResource(props['host'], int(props['port']), props['udid'],  
                props['is_simulator'], props['name'], props['version'], props['csst_uri'], props['id'])

        self._base_url = self._device_resource.driver_url
        self._ws_uri = self._device_resource.ws_uri
        self._host = RPCClientProxy('/'.join([self._base_url, 'host/']), self._ws_uri, allow_none=True, encoding=Encoding)
        self._device_udid = self._device_resource.udid
        self._device_name = self._device_resource.name
        self._device_ios = self._device_resource.version
        if isinstance(self._device_resource.is_simulator, bool):
            self._device_simulator = self._device_resource.is_simulator
        else:
            self._device_simulator = self._device_resource.is_simulator == str(True)
        
        if self._device_simulator:
            self._host.start_simulator(self._device_udid)
        self._app_started = False
        Device.Devices.append(self)
        url = '/'.join([self._base_url, 'device', '%s/' % self._device_udid])
        self._driver = RPCClientProxy(url, self._ws_uri, allow_none=True, encoding=Encoding)
        self._keyboard = Keyboard(self)
        logger.info('[%s] Device - Connect - %s - %s (%s)' % (datetime.datetime.fromtimestamp(time.time()), self.name, self.udid, self.ios_version))
        # 申请设备成功后，对弹窗进行处理
        rule = settings.get('QT4I_ALERT_DISMISS', DEFAULT_ALERT_RULE)
        if rule :
            try :
                self._dismiss_alert(rule)
            except :
                logger.exception('dismiss alert %s' % rule)

    @property
    def driver(self):
        '''设备所在的driver
        
        :rtype: RPCClientProxy
        '''
        return self._driver

    @property
    def udid(self):
        '''设备的udid
        
        :rtype: str
        '''
        result = self._device_udid
        if PY2 and isinstance(result, six.text_type):
            return result.encode(Encoding)
        return result
    
    @property
    def name(self):
        '''设备名
        
        :rtype: str
        '''
        result = self._device_name
        if PY2 and isinstance(result, six.text_type):
            return result.encode(Encoding)
        return result
        
    @property
    def ios_version(self):
        '''iOS版本
        
        :rtype: str
        '''
        result = self._device_ios
        if PY2 and isinstance(result, six.text_type):
            return result.encode(Encoding)
        return result

    @property
    def simulator(self):
        '''是否模拟器
        
        :rtype: boolean
        '''
        return self._device_simulator
    
    @property
    def rect(self):
        '''屏幕大小
        
        :rtype: Rectangle
        '''
        rect   = self._driver.device.get_rect()
        origin = rect['origin']
        size   = rect['size']
        return Rectangle(origin['x'], origin['y'], origin['x'] + size['width'], origin['y'] + size['height'])

    @property
    def keyboard(self):
        '''获取键盘
        '''
        return self._keyboard
    
    def start_app(self, bundle_id, app_params, env, trace_template=None, trace_output=None, retry=5, timeout=55):
        '''启动APP
        
        :param bundle_id: APP Bundle ID
        :type bundle_id: str
        :param app_params: app启动参数
        :type app_params: dict
        :param env: app的环境变量
        :type env: dict
        :param trace_template: 专项使用trace_template路径，或已配置的项
        :type trace_template: str
        :param trace_output: 专项使用trace_output路径
        :type trace_output: str
        :param retry: 重试次数（建议大于等于2次）
        :type retry: int
        :param timeout: 单次启动超时（秒）
        :type timeout: int
        :param instruments_timeout: 闲置超时（秒）
        :type instruments_timeout: int
        :rtype: boolean
        '''
        if not self._app_started:
            if self._driver.device.start_app(bundle_id, app_params, env, trace_template, trace_output, retry, timeout):
                self._app_started = True
                self.bundle_id = bundle_id
        elif 'app_name' in app_params and self.get_foreground_app_name() != app_params['app_name']:
            try:
                elements = self._driver.element.find_elements(app_params['app_name'], 2, 0.05, 'id')
                element_id = elements['elements'][-1]['element'] 
                self._driver.element.click(element_id)
                self.bundle_id = bundle_id
                self._app_started = True
            except:
                raise Exception('未找到名为[%s]的App' % app_params['app_name'])
        return self._app_started == True
    
    def stop_app(self):
        '''终止APP
        
        :rtype: boolean
        '''
        if self._app_started:
            if self._driver.device.stop_app(self.bundle_id):
                self._app_started = False
        return self._app_started == False

    def release(self):
        '''释放设备
        '''
        try:
            self.stop_app()
            self._driver.web.release()
            if hasattr(self, 'nlc_flag') and self.nlc_flag: #恢复网络设置
                from qt4i.app import NLCType
                self.switch_network(5, NLCType.NONE)
            if hasattr(self, 'wifi') and self.wifi: #关闭host代理
                self.reset_host_proxy()            
        except Exception:
            traceback.print_exc()
        finally:
            try:
                self._test_resources.release_resource("ios", self._device_resource.resource_id)
            except:
                pass
            Device.Devices.remove(self)
            logger.info('[%s] Device - Release - %s - %s (%s)' % (datetime.datetime.fromtimestamp(time.time()), self.name, self.udid, self.ios_version))

    def get_foreground_app_name(self):
        '''获取前台app的名称
        '''
        return self._driver.device.get_foreground_app_name()
    
    def get_foreground_app_pid(self):
        '''获取前台app的PID
        '''
        return self._driver.device.get_foreground_app_pid()

    def _check_app_started(self):
        '''检测APP是否已启动
        
        :raises: Exception
        '''
        if not self._app_started:
            raise Exception('app not started')
    
    def screenshot(self, image_path=None):
        '''截屏，返回元组：截屏是否成功，图片保存路径
        
        :param image_path: 截屏图片的存放路径
        :type image_path: str
        :rtype: tuple (boolean, str)
        '''
        try:
            base64_img = self._driver.device.capture_screen()
            if not image_path:
                image_path = os.path.join(QT4i_LOGS_PATH, "p%s_%s.png" %(os.getpid(), uuid.uuid1()))
            with open(os.path.abspath(image_path), "wb") as fd:
                fd.write(base64.decodestring(base64_img))
            return (os.path.isfile(image_path), image_path)
        except:
            logger.error('screenshot failed: %s' % traceback.format_exc())
            return (False, "")
    
    def print_uitree(self, need_back = False):
        '''打印界面树
        
        :param need_back: 是否需要返回UI Tree
        :type need_back: boolean
        :return: 控件树
        :rtype: dict or None
        '''
        self._check_app_started()
        _ui_tree = self._driver.device.get_element_tree()
        def _print(_tree, _spaces='', _indent='|---'):
            _line = _spaces + '{  ' + ',   '.join([
                'classname: "%s"' % _tree['classname'],
                'label: %s' % ('"%s"' % _tree['label'] if _tree['label'] else 'null'),
                'name: %s' % ('"%s"' % _tree['name' ] if _tree['name' ] else 'null'),
                'value: %s' % ('"%s"' % _tree['value'] if _tree['value'] else 'null'),
                'visible: %s' % ('true' if _tree['visible'] else 'false'),
                'enabled: %s' % ('true' if _tree['enabled'] else 'false'),
                'rect: %s' % _tree['rect'] if 'rect' in _tree else '']) + '  }'
            print(_line)
            for _child in _tree['children']: _print(_child, _spaces + _indent, _indent)
        _print(_ui_tree)
        if need_back:
            return _ui_tree
    
    def click(self, x=0.5, y=0.5):
        '''点击屏幕
        
        :param x: 横向坐标（从左向右，屏幕百分比）
        :type x: float
        :param y: 纵向坐标（从上向下，屏幕百分比）
        :type y: float
        '''
        self._check_app_started()
        self._driver.device.click(x, y)
    
    def click2(self, element):
        '''基于控件坐标点击屏幕（用于直接点击控件无效的场景,尽量少用）
        
        :param element: 控件对象
        :type element: Element
        '''
        self._check_app_started()
        element_rect = element.rect
        device_rect = self.rect
        x = (element_rect.left + element_rect.right) / (2.0 * device_rect.width)
        y = (element_rect.top + element_rect.bottom) / (2.0 * device_rect.height)
        self._driver.device.click(x, y)
        
    def long_click(self, x, y, duration=3):
        '''长按屏幕
        
        :param x: 横向坐标（从左向右，屏幕百分比）
        :type x: float
        :param y: 纵向坐标（从上向下，屏幕百分比）
        :type y: float
        :param duration: 持续时间（秒）
        :type duration: float
        '''
        self._check_app_started()
        self._driver.device.long_click(x, y, duration)
        
    def double_click(self, x, y):
        '''双击屏幕
        
        :param x: 横向坐标（从左向右计算，屏幕百分比）
        :type x: float
        :param y: 纵向坐标（从上向下计算，屏幕百分比）
        :type y: float
        '''
        self._check_app_started()
        self._driver.device.double_click(x, y)
    
    def _drag_from_to_for_duration(self, from_point, to_point, duration=0.5, repeat=1, interval=0):
        '''全屏拖拽
        
        :attention: 如有过场动画，需要等待动画完毕
        :param from_point  : 起始坐标偏移百分比
        :type from_point   : dict : { x: 0.5, y: 0.8 }
        :param to_point    : 结束坐标偏移百分比
        :type to_point     : dict : { x: 0.5, y: 0.1 }
        :param duration    : 持续时间（秒）
        :type duration     : float
        :param repeat      : 重复该操作
        :type repeat       : int
        :param interval    : 重复该操作的间隙时间（秒）
        :type interval     : float
        '''
        self._check_app_started()
        raise NotImplementedError
    
    def drag(self, from_x=0.9, from_y=0.5, to_x=0.1, to_y=0.5, duration=0.5):
        '''回避屏幕边缘，全屏拖拽（默认在屏幕中央从右向左拖拽）
        
        :param from_x: 起点 x偏移百分比(从左至右为0.0至1.0)
        :type from_x: float
        :param from_y: 起点 y偏移百分比(从上至下为0.0至1.0)
        :type from_y: float
        :param to_x: 终点 x偏移百分比(从左至右为0.0至1.0)
        :type to_x: float
        :param to_y: 终点 y偏移百分比(从上至下为0.0至1.0)
        :type to_y: float
        :param duration: 持续时间（秒）
        :type duration: float
        '''
        self._check_app_started()
        self._driver.device.drag(from_x, from_y, to_x, to_y, duration)
    
    def drag2(self, direct=EnumDirect.Left):
        '''回避屏幕边缘，全屏在屏幕中央拖拽
        
        :param direct: 拖拽的方向
        :type  direct: EnumDirect.Left|EnumDirect.Right|EnumDirect.Up|EnumDirect.Down
        '''
        self._check_app_started()
        if direct == EnumDirect.Left  : self._driver.device.drag(0.5, 0.5, 0.1, 0.5, 0.5)
        if direct == EnumDirect.Right : self._driver.device.drag(0.5, 0.5, 0.9, 0.5, 0.5)
        if direct == EnumDirect.Up    : self._driver.device.drag(0.5, 0.5, 0.5, 0.1, 0.5)
        if direct == EnumDirect.Down  : self._driver.device.drag(0.5, 0.5, 0.5, 0.9, 0.5)
    
    def flick(self, from_x=0.9, from_y=0.5, to_x=0.1, to_y=0.5):
        '''回避屏幕边缘，全屏滑动/拂去（默认从右向左滑动/拂去）
        该接口比drag的滑动速度快，如果滚动距离大，建议用此接口
        
        :param from_x: 起点 x偏移百分比（从左至右为0.0至1.0）
        :type from_x: float
        :param from_y: 起点 y偏移百分比（从上至下为0.0至1.0）
        :type from_y: float
        :param to_x: 终点 x偏移百分比（从左至右为0.0至1.0）
        :type to_x: float
        :param to_y: 终点 y偏移百分比（从上至下为0.0至1.0）
        :type to_y: float
        '''
        self._check_app_started()
        self._driver.device.drag(from_x, from_y, to_x, to_y, 0)
    
    def flick2(self, direct=EnumDirect.Left):
        '''回避屏幕边缘，全屏在屏幕中央滑动/拂去
        
        :param direct: 滑动/拂去的方向
        :type  direct: EnumDirect.Left|EnumDirect.Right|EnumDirect.Up|EnumDirect.Down
        '''
        self._check_app_started()
        if direct == EnumDirect.Left  : self._driver.device.drag(0.5, 0.5, 0.1, 0.5, 0)
        if direct == EnumDirect.Right : self._driver.device.drag(0.4, 0.5, 0.9, 0.5, 0)
        if direct == EnumDirect.Up    : self._driver.device.drag(0.5, 0.5, 0.5, 0.1, 0)
        if direct == EnumDirect.Down  : self._driver.device.drag(0.5, 0.5, 0.5, 0.9, 0)

    def flick3(self, from_x=0.5, from_y=0.8, to_x=0.5, to_y=0.2, repeat=1, interval=0.5, velocity=1000):
        '''全屏连续滑动（默认从下向滑动）
        该接口比flick2的滑动速度快，适用于性能测试
        
        :param from_x: 起点 x偏移百分比（从左至右为0.0至1.0）
        :type from_x: float
        :param from_y: 起点 y偏移百分比（从上至下为0.0至1.0）
        :type from_y: float
        :param to_x: 终点 x偏移百分比（从左至右为0.0至1.0）
        :type to_x: float
        :param to_y: 终点 y偏移百分比（从上至下为0.0至1.0）
        :type to_y: float
        :param repeat: 滑动的次数
        :type repeat: int
        :param interval: 滑动的间隔时间（秒）
        :type interval: double
        :param velocity: 滑动的速度
        :type velocity: double
        '''
        self._check_app_started()
        self._driver.device.drag(from_x, from_y, to_x, to_y, 0, repeat, interval, velocity)
        
    def deactivate_app_for_duration(self, seconds=3):
        '''将App置于后台一定时间
        
        :param seconds: 秒,若为-1，则模拟按Home键的效果，也即app切到后台后必须点击app才能再次唤起
        :type seconds: int
        :return: boolean
        '''
        return self._driver.device.background_app(seconds)
       
    def install(self, app_path):
        '''安装应用程序
        
        :param app_path: ipa或app安装包的路径(注意：真机和模拟器的安装包互不兼容)
        :type app_path: str
        :rtype: boolean
        '''
        begin_time = time.time()
        result = self._driver.device.install(app_path) 
        count_time = time.time() - begin_time
        logger.info("安装被测应用耗时：%.3fs" % count_time)
        return result
    
    def uninstall(self, bundle_id):
        '''卸载应用程序
        
        :param bundle_id: APP的bundle_id，例如：com.tencent.qq.dailybuild.test
        :type bundle_id: str
        :rtype: boolean
        '''
        return self._driver.device.uninstall(bundle_id)
    
    def get_crash_log(self, procname):
        '''获取指定进程的最新的crash日志
        
        :param proc_name: app的进程名，可通过xcode查看
        :type proc_name: str  
        :return: crash日志路径
        :rtype: string or None
        '''
        crash_log = self._driver.device.get_crash_log(procname)
        if crash_log:
            crash_log_path = os.path.join(QT4i_LOGS_PATH, "%s_%s.crash" %(procname, uuid.uuid1()))
            with open(os.path.abspath(crash_log_path), "wb") as fd:
                fd.write(crash_log)
            return crash_log_path
        else:
            return None
    
    def pull_file(self, bundle_id, remotepath, localpath=None, is_dir=False, is_delete=True):
        '''拷贝手机中sandbox指定目录的文件到Mac本地
        
        :param bundle_id: app的bundle id
        :type bundle_id: str   
        :param remotepath: sandbox上的目录或者文件，例如：/Library/Caches/test/
        :type remotepath: str        
        :param localpath: 本地的目录
        :type localpath: str
        :param is_dir: remotepath是否为目录，默认为单个文件
        :type is_dir: boolean
        :return: 拷贝到本地的文件列表
        :rtype: list or None
        '''
        if localpath is None:
            localpath = QT4i_LOGS_PATH
        filepaths = []
        if self._device_resource.host != DEFAULT_ADDR:
            files = self._driver.device.pull_file(bundle_id, remotepath, '/tmp', is_dir, is_delete)
            for f in files:
                filepath = os.path.join(localpath, os.path.basename(f))
                with open(filepath, "wb") as fd:
                    data = self._host.pull_file_data(f, 0)
                    index = 1
                    while data != None:
                        fd.write(base64.decodestring(data))
                        data = self._host.pull_file_data(f, index)
                        index += 1
                filepaths.append(filepath)
        else:
            filepaths = self._driver.device.pull_file(bundle_id, remotepath, localpath, is_dir, is_delete)
        return filepaths
    
    copy_to_local = pull_file

    def push_file(self, bundle_id, localpath, remotepath):
        '''拷贝Mac本地文件到手机中sandbox的指定目录地
        
        :param bundle_id: app的bundle id
        :type bundle_id: str   
        :param localpath: 文件路径，支持本地文件路径和http路径
        :type localpath: str
        :param remotepath: iPhone上的目录或者文件，例如：/Documents/
        :type remotepath: str
        :rtype: boolean
        '''
        if localpath.startswith('http'):
            return self._driver.device.download_file_and_push(bundle_id, localpath, remotepath)
        else:
            if self._device_resource.host != DEFAULT_ADDR:
                with open(localpath, "rb") as fd:
                    data = base64.b64encode(fd.read())
                    filepath = os.path.join('/tmp', os.path.basename(localpath))
                    self._host.push_file_data(data, filepath)
                    localpath = filepath
            return self._driver.device.push_file(bundle_id, localpath, remotepath)

    def list_files(self, bundle_id, file_path):
        '''列出手机上app中的文件或者目录
        
        :param bundle_id: app的bundle id
        :type bundle_id: str   
        :param file_path: sandbox上的目录或者文件，例如：/Library/Caches/test/
        :type file_path: str
        ''' 
        return self._driver.device.list_files(bundle_id, file_path)

    def download_file(self, file_url, dst_path):
        '''从网上下载指定文件到本地
        
        :param file_url: 文件的url路径，支持http和https路径, 需要对app插桩
        :type file_url: str
        :param dst_path: 文件在手机上的存储路径，例如：/Documents
        :type dst_path: str
        '''  
        method = 'createFile:withPath:size:'
        params = [file_url, dst_path, 1]
        return self.call_qt4i_stub(method, params)

    def remove_files(self, bundle_id, file_path):
        '''删除手机上app中的文件或者目录(主要用于app的日志或者缓存的清理)
        
        :param bundle_id: app的bundle id
        :type bundle_id: str   
        :param file_path: sandbox上的目录或者文件，例如：/Library/Caches/test/
        :type file_path: str
        ''' 
        self._driver.device.remove_files(bundle_id, file_path)       

    remove_file = remove_files

    def reboot(self):
        '''重启手机
        '''
        self._driver.device.reboot()
        
        
    def get_driver_log(self, start_time=None):
        '''获取driver日志
        :param start_time 用例执行的开始时间
        :type str
        
        try-catch 说明：为了兼容client以及server端，对取失败用例日志的操作（后续接口更新可删除）
        '''
        try:
            return self._driver.device.get_driver_log(start_time)
        except:
            return self._driver.device.get_driver_log()
        
    def get_log(self, start_time=None):
        '''获取交互日志
        :param start_time 用例执行的开始时间
        :type str
        
        try-catch 说明：为了兼容client以及server端，对取失败用例日志的操作（后续接口更新可删除）
        '''
        try:
            return self._driver.device.get_log(start_time)
        except:
            return self._driver.device.get_log()
        
    def get_syslog(self, watchtime, process_name=None):
        '''获取手机系统日志
        '''
        return self._driver.device.get_syslog(watchtime, process_name)
        
    def cleanup_log(self):
        '''清理交互日志
        '''
        self._driver.device.cleanup_log()
        logger.info('[%s] Device - Clean Up Logger - %s - %s (%s)' % (datetime.datetime.fromtimestamp(time.time()), self.name, self.udid, self.ios_version))
    
    def get_app_list(self, app_type="user"):
        '''获取设备上的app列表
        
        :param app_type: app的类型(user/system/all)
        :type app_type: str 
        :return: list
        :rtype: app列表，例如: [{'com.tencent.demo': 'Demo'}]
        '''   
        return self._driver.device.get_app_list(app_type) 
    
    def lock(self):
        '''锁定设备（灭屏）
        '''
        self._driver.device.lock()
        
    def unlock(self):
        '''解锁设备
        '''
        self._driver.device.unlock()    
        
    def _volume(self, cmd):
        '''音量调节
        
        :param cmd: 音量调节命令
        :type str : 'up'|'down'|'silent'
        '''
        self._driver.device.volume(cmd)
        
    def _screen_direction(self, direct):
        '''设置屏幕方向
        
        :param direct: 屏幕调节命令
        :type str : 'up'|'down'|'left'|'right'
        '''
        self._driver.device.screen_direction(direct)
        
    def _siri(self, cmd):
        '''siri交互
        
        :param cmd: 与siri交互文本
        :type str : 譬如打开一个app，直接cmd='app名'
        '''
        self._driver.device.siri(cmd)
        
    def _dismiss_alert(self, rule):
        '''弹窗处理
        
        :param rule: 弹窗处理
        :type list: rules
        '''
        self._driver.device.dismiss_alert(rule)
    
    def switch_network(self, network_type, nlc_type):
        '''实现网络切换
        
        :param network_type: 网络类型，例如：
                            0:无WIFI无xG
                            1:无WIFi有xG
                            2:有WIFI无xG
                            3:有WIFI有xG
                            4:飞行模式
        :type network_type: int
        :param nlc_type: 模拟弱网络类型 
        :type nlc_type: NLCType
        '''
        self.nlc_flag = True
        from qt4i.app import Preferences
        if self._app_started:
            self._active_app_name = self.get_foreground_app_name()
            self.deactivate_app_for_duration(-1)
        else:
            self._active_app_name = None
        pre = Preferences(self)
        pre.switch_network(network_type, nlc_type)
        if self._active_app_name:
            self.deactivate_app_for_duration(-1)
            self.start_app('', {'app_name':self._active_app_name}, None)
        else:
            self.stop_app()

    def set_host_proxy(self, server, port, wifi):
        '''设置host代理
        
        :param server: 服务器名
        :type server: str
        :param port: 端口号
        :type port: int
        :param wifi: wifi名
        :type wifi: str
        '''
        self.wifi = wifi
        from qt4i.app import Preferences
        if self._app_started:
            self._active_app_name = self.get_foreground_app_name()
            self.deactivate_app_for_duration(-1)
        else:
            self._active_app_name = None
        pre = Preferences(self)
        pre.set_host_proxy(server, port, wifi)
        if self._active_app_name:
            self.deactivate_app_for_duration(-1)
            self.start_app('', {'app_name':self._active_app_name}, None)
        else:
            self.stop_app()
    
    def reset_host_proxy(self):
        '''关闭host代理
        '''
        from qt4i.app import Preferences
        pre = Preferences(self)
        pre.reset_host_proxy()
                        
    def upload_photo(self, photo_path, album_name, cleared=True):
        '''上传照片到系统相册
        
        :param photo_path: 本地照片路径
        :type photo_path: str
        :param album_name: 系统相册名
        :type album_name: str
        :param cleared: 是否清空已有的同名相册
        :type cleared: boolean
        :rtype: boolean
        '''
        with open(photo_path, 'rb') as fd:
            img_data = base64.b64encode(fd.read())
            return self._driver.device.upload_photo(img_data, album_name, cleared)
    
    def call_qt4i_stub(self, method, params, clazz=None):
        '''QT4i Stub通用接口

        :param method: 函数名
        :type method: str
        :param params: 函数参数
        :type params: list
        :param clazz: 类名
        :type clazz: str
        :return: 插桩接口返回值
        :rtype: str
        '''
        return self._driver.device.call_qt4i_stub(method, params, clazz)

    def get_device_detail(self): 
        '''获取设备型号，颜色，内存大小等详细信息
        '''
        return self._driver.device.get_device_detail()

    def get_icon_badge(self, app_name):
        '''获取设备中app消息未读数
        '''
        if self._app_started:
            self._active_app_name = self.get_foreground_app_name()
            self.deactivate_app_for_duration(-1)
        else:
            self._active_app_name = None   
        elements = self._driver.element.find_elements(app_name, 2, 0.05, 'id')
        element_id = elements['elements'][-1]['element'] 
        value = self._driver.element.get_element_attr(element_id, 'value')
        if value == None :
            value = 0
        else :
            value = value.split(' ')
            value = value[0] 
        if self._active_app_name:
            self.deactivate_app_for_duration(-1)
            self.start_app('', {'app_name':self._active_app_name}, None)
        else:
            self.stop_app()
        return value


class Keyboard(object):
    '''键盘
    '''

    def __init__(self, device):
        self._device = device
        self._driver = device.driver

    def send_keys(self, keys):
        '''键盘输入
        
        :param keys: 要输入的字符串
        :type keys: str
        '''
        self._driver.device.send_keys(keys)


# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''异常处理器
'''

from __future__ import absolute_import, print_function

import six

from qt4i.driver.xctest.webdriverclient.exceptions import ElementNotSelectableException
from qt4i.driver.xctest.webdriverclient.exceptions import ElementNotVisibleException
from qt4i.driver.xctest.webdriverclient.exceptions import InvalidElementStateException
from qt4i.driver.xctest.webdriverclient.exceptions import InvalidSelectorException
from qt4i.driver.xctest.webdriverclient.exceptions import InvalidQPathException
from qt4i.driver.xctest.webdriverclient.exceptions import ImeNotAvailableException
from qt4i.driver.xctest.webdriverclient.exceptions import ImeActivationFailedException
from qt4i.driver.xctest.webdriverclient.exceptions import NoSuchSessionException
from qt4i.driver.xctest.webdriverclient.exceptions import NoSuchElementException
from qt4i.driver.xctest.webdriverclient.exceptions import NoSuchFrameException
from qt4i.driver.xctest.webdriverclient.exceptions import NoSuchWindowException
from qt4i.driver.xctest.webdriverclient.exceptions import StaleElementReferenceException
from qt4i.driver.xctest.webdriverclient.exceptions import UnexpectedAlertPresentException
from qt4i.driver.xctest.webdriverclient.exceptions import NoAlertPresentException
from qt4i.driver.xctest.webdriverclient.exceptions import ErrorInResponseException
from qt4i.driver.xctest.webdriverclient.exceptions import TimeoutException
from qt4i.driver.xctest.webdriverclient.exceptions import WebDriverException
from qt4i.driver.xctest.webdriverclient.exceptions import MoveTargetOutOfBoundsException
from qt4i.driver.xctest.webdriverclient.exceptions import RotationNotAllowedException
from qt4i.driver.xctest.webdriverclient.exceptions import ApplicationDeadlockException
from qt4i.driver.xctest.webdriverclient.exceptions import ApplicationCrashedException
from qt4i.driver.xctest.webdriverclient.exceptions import XCTestAgentDeadException
from qt4i.driver.xctest.webdriverclient.exceptions import SendKeysFailedException




class ErrorCode(object):
    """
    Error codes defined in the WebDriver wire protocol.
    """
    # Keep in sync with org.openqa.selenium.remote.ErrorCodes and errorcodes.h
    SUCCESS = 0
    UNSUPPORTED = [1, 'unsupported']
    NO_SUCH_SESSION = [6, 'no such session']
    NO_SUCH_ELEMENT = [7, 'no such element']
    NO_SUCH_FRAME = [8, 'no such frame']
    UNKNOWN_COMMAND = [9, 'unknown command']
    STALE_ELEMENT_REFERENCE = [10, 'stale element reference']
    ELEMENT_NOT_VISIBLE = [11, 'element not visible']
    INVALID_ELEMENT_STATE = [12, 'invalid element state']
    UNKNOWN_ERROR = [13, 'unknown error']
    ELEMENT_IS_NOT_SELECTABLE = [15, 'element not selectable']
    JAVASCRIPT_ERROR = [17, 'javascript error']
    XPATH_LOOKUP_ERROR = [19, 'invalid selector']
    TIMEOUT = [21, 'timeout']
    NO_SUCH_WINDOW = [23, 'no such window']
    INVALID_COOKIE_DOMAIN = [24, 'invalid cookie domain']
    UNABLE_TO_SET_COOKIE = [25, 'unable to set cookie']
    UNEXPECTED_ALERT_OPEN = [26, 'unexpected alert open']
    NO_ALERT_OPEN = [27, 'no such alert']
    SCRIPT_TIMEOUT = [28, 'script timeout']
    INVALID_ELEMENT_COORDINATES = [29, 'invalid element coordinates']
    IME_NOT_AVAILABLE = [30, 'ime not available']
    IME_ENGINE_ACTIVATION_FAILED = [31, 'ime engine activation failed']
    INVALID_SELECTOR = [32, 'invalid selector']
    SESSION_NOT_CREATED = [33, 'session not created']
    MOVE_TARGET_OUT_OF_BOUNDS = [34, 'move target out of bounds']
    INVALID_XPATH_SELECTOR = [51, 'invalid selector']
    INVALID_XPATH_SELECTOR_RETURN_TYPER = [52, 'invalid selector']
    INVALID_QPATH = [60, 'invalid qpath']
    METHOD_NOT_ALLOWED = [405, 'unsupported operation']
    ROTATION_NOT_ALLOWED = [777, 'rotation not allowed']
    APPLICATION_DEADLOCK_DETECTED = [888, 'application deadlock detected']
    APPLICATION_CRASHED = [889, 'application crashed']
    XCTestAgent_NOT_BACKGROUND = [890, 'XCTestAgent Dead']
    SENDKEYS_FAILED = [891, 'sendkeys failed']
    QT4I_STUB_NOT_EXIST = 900


class ErrorHandler(object):
    """
    Handles errors returned by the WebDriver server.
    """
    ERROR_IN_RESPONSE_EXCEPTIONS = [
                            ErrorInResponseException, 
                            SendKeysFailedException,
                            XCTestAgentDeadException,
                            ApplicationCrashedException,
                            ]
    
    def check_response(self, response):
        """
        Checks that a JSON response from the WebDriver does not have an error.

        :Args:
         - response - The JSON response from the WebDriver server as a dictionary
           object.

        :Raises: If the response contains an error message.
        """
        status = response.get('status', None)
        if status is None or status == ErrorCode.SUCCESS \
            or status == ErrorCode.QT4I_STUB_NOT_EXIST:
            return

        value = None
        message = response.get("message", "")
        screen = response.get("screen", "")
        stacktrace = None
        
        exception_class = ErrorInResponseException
        if status in ErrorCode.NO_SUCH_SESSION \
                or status in ErrorCode.SESSION_NOT_CREATED:
            exception_class = NoSuchSessionException
        elif status in ErrorCode.NO_SUCH_ELEMENT:
            exception_class = NoSuchElementException
        elif status in ErrorCode.NO_SUCH_FRAME:
            exception_class = NoSuchFrameException
        elif status in ErrorCode.NO_SUCH_WINDOW:
            exception_class = NoSuchWindowException
        elif status in ErrorCode.STALE_ELEMENT_REFERENCE:
            exception_class = StaleElementReferenceException
        elif status in ErrorCode.ELEMENT_NOT_VISIBLE:
            exception_class = ElementNotVisibleException
        elif status in ErrorCode.INVALID_ELEMENT_STATE:
            exception_class = InvalidElementStateException
        elif status in ErrorCode.INVALID_SELECTOR \
                or status in ErrorCode.INVALID_XPATH_SELECTOR \
                or status in ErrorCode.INVALID_XPATH_SELECTOR_RETURN_TYPER:
            exception_class = InvalidSelectorException
        elif status in ErrorCode.INVALID_QPATH:
            exception_class = InvalidQPathException
        elif status in ErrorCode.ELEMENT_IS_NOT_SELECTABLE:
            exception_class = ElementNotSelectableException
        elif status in ErrorCode.INVALID_COOKIE_DOMAIN:
            exception_class = WebDriverException
        elif status in ErrorCode.UNABLE_TO_SET_COOKIE:
            exception_class = WebDriverException
        elif status in ErrorCode.TIMEOUT:
            exception_class = TimeoutException
        elif status in ErrorCode.SCRIPT_TIMEOUT:
            exception_class = TimeoutException
        elif status in ErrorCode.UNKNOWN_ERROR:
            exception_class = WebDriverException
        elif status in ErrorCode.UNEXPECTED_ALERT_OPEN:
            exception_class = UnexpectedAlertPresentException
        elif status in ErrorCode.NO_ALERT_OPEN:
            exception_class = NoAlertPresentException
        elif status in ErrorCode.IME_NOT_AVAILABLE:
            exception_class = ImeNotAvailableException
        elif status in ErrorCode.IME_ENGINE_ACTIVATION_FAILED:
            exception_class = ImeActivationFailedException
        elif status in ErrorCode.MOVE_TARGET_OUT_OF_BOUNDS:
            exception_class = MoveTargetOutOfBoundsException
        elif status in ErrorCode.ROTATION_NOT_ALLOWED:
            exception_class = RotationNotAllowedException
        elif status in ErrorCode.APPLICATION_DEADLOCK_DETECTED:
            exception_class = ApplicationDeadlockException
        elif status in ErrorCode.APPLICATION_CRASHED:
            exception_class = ApplicationCrashedException
        elif status in ErrorCode.XCTestAgent_NOT_BACKGROUND:
            exception_class = XCTestAgentDeadException
        elif status in ErrorCode.SENDKEYS_FAILED:
            exception_class = SendKeysFailedException
        else:
            exception_class = WebDriverException
        
        if isinstance(status, int):
            value_json = response.get('value', None)
            if value_json and isinstance(value_json, six.string_types):
                try:
                    import json
                    value = json.loads(value_json)
                except:
                    pass
                else:
                    status = value['error']
                    message = value['message']
                    raise exception_class(response, message)

        value = response['value']
        if isinstance(value, six.string_types):
            if exception_class in self.ERROR_IN_RESPONSE_EXCEPTIONS:
                raise exception_class(response, value)
            raise exception_class(value)
        message = ''
        if 'message' in value:
            message = value['message']

        screen = None
        if 'screen' in value:
            screen = value['screen']

        stacktrace = None
        if 'stackTrace' in value and value['stackTrace']:
            stacktrace = []
            try:
                for frame in value['stackTrace']:
                    line = self._value_or_default(frame, 'lineNumber', '')
                    file = self._value_or_default(frame, 'fileName', '<anonymous>')
                    if line:
                        file = "%s:%s" % (file, line)
                    meth = self._value_or_default(frame, 'methodName', '<anonymous>')
                    if 'className' in frame:
                        meth = "%s.%s" % (frame['className'], meth)
                    msg = "    at %s (%s)"
                    msg = msg % (meth, file)
                    stacktrace.append(msg)
            except TypeError:
                pass
        if exception_class in self.ERROR_IN_RESPONSE_EXCEPTIONS:
            raise exception_class(response, message)
        elif exception_class == UnexpectedAlertPresentException and 'alert' in value:
            raise exception_class(message, screen, stacktrace, value['alert'].get('text'))
        raise exception_class(message, screen, stacktrace)

    def _value_or_default(self, obj, key, default):
        return obj[key] if key in obj else default



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''控件定位策略
'''

class By(object):
    """
    Set of supported locator strategies.
    """

    ID = "id"
    NAME = "name"
    QPATH = "qpath"
    XPATH = "xpath"
    TAG_NAME = "tag name"
    LINK_TEXT = "link text"
    CLASS_NAME = "class name"
    PREDICATE_STRING = "predicate string"
    PARTIAL_LINK_TEXT = "partial link text"

    @classmethod
    def is_valid(cls, by):
        for attr in dir(cls):
            if by == getattr(cls, attr):
                return True
        return False


# coding: UTF-8
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''用于与Server通信（xmlrpc）
'''
from __future__ import absolute_import, print_function

import os, re, sys, socket, six
from qt4i.driver.rpc import RPCClientProxy
# -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*-

class Args(dict):
    '''@desc: 命令行参数解析为字典（支持格式:  -key "value" -key=value --key "value" --key=value）
    '''
    RegPattern1 = '^\-{1,2}\w+\=.+$'
    RegPattern2 = '^\-{1,2}\w+$'
    RegPatternAll = '|'.join([RegPattern1, RegPattern2])
    
    def __init__(self, keys_to_lower=True):
        dict.__init__(self)
        self.keys_to_lower = keys_to_lower
        self.parse()
    
    def __convert__(self, item):
        if re.match('^\d+\.\d+$', item)                          : return float(item)
        if re.match('^\d+$', item)                               : return int(item)
        if re.match('|'.join(['^True$', '^False$']), item, re.I) : return bool(item)
        if re.match('|'.join(['^None$', '^NULL$']), item, re.I)  : return None
        return item
    
    def update(self, _dict):
        if self.keys_to_lower:
            for (key, value) in six.iteritems(_dict):
                self[key.lower()] = value
        else:super(self.__class__, self).update(_dict)
    
    def parse(self):
        for i, item in enumerate(sys.argv):
            if re.match(self.RegPatternAll, item):
                if re.match(self.RegPattern1, item):
                    item = re.sub('^-+', '', item)
                    key = re.sub('=.+$', '', item)
                    value = re.sub('^\w+=', '', item)
                    if self.keys_to_lower : key = key.lower()
                    self.update({key:self.__convert__(value)})
                    continue
                if re.match(self.RegPattern2, item):
                    key = re.sub('^-+', '', item)
                    value = sys.argv[i + 1]
                    if self.keys_to_lower : key = key.lower()
                    self.update({key:self.__convert__(value)})
                    continue

# -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*-

class Rpc(object):
    
    def __init__(self, xmlrpc_uri, rpc_method, device_udid, result, timeout):
        self.xmlrpc_uri = xmlrpc_uri
        self.rpc_method = rpc_method
        self.result = result
        self.timeout = timeout
        self.rpc = RPCClientProxy(self.xmlrpc_uri, allow_none=True, encoding='UTF-8')
        self.rpc_working = RPCClientProxy(''.join([self.xmlrpc_uri.split("device")[0],'host/']), allow_none=True, encoding='UTF-8')
        socket.setdefaulttimeout(self.timeout + 3)
    
    def rpc_is_working(self):
        try: return self.rpc_working.echo()
        except: return False
    
    def send_result(self):
        if self.rpc_is_working(): self.rpc.ins.send_result(self.result)
    
    def send_result_and_get_next(self):
        # //<-- return  : {"id": 1, "result": "JSON", "error": null}
        # //--> command : {"id": 1, "method": "echo", "params": ["Hello JSON"]}
        if self.rpc_is_working(): return self.rpc.ins.send_result_and_get_next(self.result, self.timeout)
        return '{"id": -1, "method": "release"}'
    
    def rpc_exec(self):
        if self.rpc_method == 'send_result_and_get_next' : return self.send_result_and_get_next()
        if self.rpc_method == 'send_result'              : self.send_result()


if __name__ == '__main__':
    try: sys.stdout.write(Rpc(**Args()).rpc_exec() or '') 
    except Exception as e:
        with open(os.path.join(os.path.dirname(__file__), '_cmd_fetch_delegate.error.log'), 'w') as fd:
            fd.write(repr(e))
        sys.stdout.write(repr(e))
        raise e



# -*- coding:utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''RPC Server Host Driver
'''

from __future__ import absolute_import, print_function

import os
import base64
from qt4i.driver.rpc import rpc_method
from qt4i.driver.rpc import RPCEndpoint
from qt4i.driver.tools.dt import DT
from qt4i.driver.xctest.agent import XCUITestAgentManager

BUFFER_SIZE = 1024*1024*100


class RPCServerHost(RPCEndpoint):
    """RPC Server Host API
    """
    agent_manager = XCUITestAgentManager()
    
    def __init__(self, rpc_server):
        self.rpc_server = rpc_server
        RPCEndpoint.__init__(self)
    
    @rpc_method
    def pull_file_data(self, filepath, seek_index, buffer_size=BUFFER_SIZE):
        '''copy file from rpc server
        
        :param filepath: file path of rpc server
        :type filename: str
        :returns: six.xmlrpc_client.Binary
        '''
        if os.path.exists(filepath):
            with open(filepath, "rb") as fd:
                seek = seek_index * buffer_size
                if seek <= os.path.getsize(filepath):
                    fd.seek(seek)
                    data = base64.b64encode(fd.read(buffer_size))
                    return data
                return None
        else:
            raise Exception('file(%s) does not exist' % filepath)

    @rpc_method
    def push_file_data(self, data, filepath):
        '''copy file to rpc server
        
        :param data: file data
        :type data: str
        :param filepath: file path of rpc server
        :type filepath: str
        :returns: boolean
        '''
        if os.path.exists(filepath):
            os.remove(filepath)
        with open(filepath, "wb") as fd:
            fd.write(base64.decodestring(data))
        return os.path.isfile(filepath)
  
    @rpc_method
    def list_devices(self):
        '''list all devices of rpc server host
        
        :return: list
        '''
        return DT().get_devices()
    
    @rpc_method
    def list_real_devices(self):
        '''list all real iOS devices of rpc server host
        
        :returns: list
        '''
        return DT().get_real_devices()
    
    @rpc_method
    def list_simulators(self):
        '''list all iOS simulators of rpc server host
        
        :returns: list
        '''
        return DT().get_simulators()
    
    @rpc_method
    def start_simulator(self, udid=None):
        '''start iOS simulator of rpc server host
        
        :param udid: udid of iOS simulator
        :type udid: str
        :returns: boolean
        '''
        return DT().start_simulator(udid)
    
    @rpc_method
    def get_device_by_udid(self, udid):
        '''inquire device by udid
        
        :param udid: udid of device
        :type udid: str
        :returns: dict or None
        '''
        return DT().get_device_by_udid(udid)
    
    @rpc_method
    def echo(self):
        '''test xmlrpc server started
        
        :returns: boolean
        '''
        return True
    
    @rpc_method
    def stop_all_agents(self):
        '''
        stop all agents
        '''
        XCUITestAgentManager.stop_all_agents()



# -*- coding:utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''iOS真机抓包工具
'''

import re
import subprocess
import thread
import time
import traceback
from six import PY3
    
def start_rvictl(udid, sudo_pwd, pcap_path='dump.pcap'):
    try:
        result = False
        cmd = 'rvictl -s %s' % udid
        start_info = subprocess.check_output(cmd.split(' '), stderr = subprocess.STDOUT)
        if PY3:
            start_info = start_info.decode()
        start_info =start_info.strip('\n')
        matched = re.match(r'Starting device \w+ \[SUCCEEDED\] with interface (rvi\d+)', start_info)
        if matched:
            interface = matched.group(1)
            tcpdump = 'tcpdump -i %s -n -s 0 tcp -w %s' % (interface, pcap_path)
            thread.start_new_thread(_sudo, (sudo_pwd, tcpdump))
            time.sleep(1)
            result = True
    except subprocess.CalledProcessError:
        traceback.print_exc()
    return result
    

def stop_rvictl(udid, sudo_pwd):
    try:
        result = False
        rvi_active = subprocess.check_output('rvictl -l'.split(' '), stderr = subprocess.STDOUT)
        if PY3:
            rvi_active = rvi_active.decode()
        interface = None
        for line in rvi_active.split('\n'):
            if udid in line:
                interface = re.match(r'.+(rvi\d+)', line).group(1)
                break
        grep = subprocess.check_output("ps ax | grep tcpdump | grep -v grep", shell=True)
        if PY3:
            grep = grep.decode()
        for line in grep.split('\n'):
            if 'tcpdump' in line and interface in line:
                pid = line.split(' ')[1]
                stop_tcpdump = 'kill -9 %s' % pid
                _sudo(sudo_pwd, stop_tcpdump)
        stop_rvi = 'rvictl -x %s' % udid
        info = subprocess.check_output(stop_rvi.split(' '), stderr = subprocess.STDOUT)
        if PY3:
            info = info.decode()
        matched = re.match(r'Stopping device \w+ \[SUCCEEDED\]', info)
        result = not matched 
    except subprocess.CalledProcessError:
        traceback.print_exc()
    return result


def _sudo(sudo_pwd, cmd):
    proc = subprocess.Popen(['sudo', '-S'] + cmd.split(), stdin=subprocess.PIPE, 
            stderr=subprocess.PIPE, universal_newlines=True)
    proc.communicate(sudo_pwd + '\n')
